local JSON = require("libs.json.JSON")
local Vector = require("libs.hump.vector")
local Map = Class("Map")

local sqrt, floor = math.sqrt, math.floor
local sin, cos, tan = math.sin, math.cos, math.tan

-- Helper functions
local function findNextSquareBase(n)
	local baseN = sqrt(n)
	local floorN = floor(baseN)
	if(baseN == floorN) then
		return floorN
	else
		return floorN + 1
	end
end
-- End helper functions

function Map.fromJSON(jsonString)
	local json = JSON:decode(jsonString)
	local map = Map:new()

	-- Tileset things
	map.tileset = load:image(json.tileset)
	map.tileW, map.tileH = json.tileW, json.tileH
	map.spriteW, map.spriteH = json.spriteW, json.spriteH

	-- Quad things
	local attributes = {}
	map.refW, map.refH = map.tileset:getDimensions()
	for _, quad in ipairs(json.quads) do
		map.quads[quad.char] = love.graphics.newQuad(quad.x * map.tileW,
													 quad.y * map.tileH,
													 map.tileW, map.tileH,
													 map.refW, map.refH)
		map.minimapIcons[quad.char] = quad.icon
		attributes[quad.char] = quad.attributes
	end

	-- Get layers
	map.layers = #json.layout
	-- Initialize layers
	for i=1, map.layers do
		map.layer[i] = {}
		-- Get layer width and height
		local width = #(json.layout[i][1]:match("[^\n]+"))
		local height = #json.layout[i]
		map.layer[i].width = width
		map.layer[i].height = height
		-- Determine max width and height
		map.width = (width > map.width) and width or map.width
		map.height = (height > map.height) and height or map.height

		-- Initialize tiles per layer
		map.layer[i].tiles = {}
		for x=0, width-1 do
			map.layer[i].tiles[x] = {}
		end
		for y, row in ipairs(json.layout[i]) do
			local x = 0
			y = y - 1
			for tile in row:gmatch(".") do
				map.layer[i].tiles[x][y] = {}
				map.layer[i].tiles[x][y].char = tile
				map.layer[i].tiles[x][y].attributes = {}
				map.layer[i].tiles[x][y].entities = {}
				map.layer[i].tiles[x][y].entityCount = 0

				if(attributes[tile]) then
					for _, attrib in pairs(attributes[tile]) do
						map.layer[i].tiles[x][y].attributes[attrib] = true
					end
				end

				x = x + 1
			end
		end
	end

	-- Get misc layer
	local miscWidth =  #(json.misc[1])
	for x=0, miscWidth-1 do
		map.misc[x] = {}
	end
	for y, row in ipairs(json.misc) do
		local x = 0
		y = y - 1
		for tile in row:gmatch(".") do
			if(tile ~= " ") then
				map.misc[tile] = {}
				map.misc[tile].x = x
				map.misc[tile].y = y
			end

			x = x + 1
		end
	end

	return map
end

function Map:init()
	-- List of entities on the map
	self.entities = {}

	-- Layer count
	self.layers = 0
	-- Layer array
	self.layer = {}
	self.floorLayer = 0
	self.raycastLayer = 0

	-- Misc information
	self.misc = {}

	-- Max width, Max height
	self.width = 0
	self.height = 0

	-- Minimap
	self.minimapIcons = {}
	self.minimapScale = 0

	-- Floorcasting
	self.floorImage = nil

	-- Raycasting
	self.raycaster = nil

	-- Spawner
	self.spawner = nil

	-- Internal Map Data
	self.tileset = nil
	self.quads = {}
	self.tileW, self.tileH = 0, 0
	self.spriteW, self.spriteH = 0, 0
	self.refW, refH = 0, 0
	self.eid = 0
end

function Map:initFloor()
	local layer = self.layer[self.floorLayer]
	local tileW, tileH = self.tileW, self.tileH

	-- Create floor canvas
	local floorWidth = self.width * self.tileW
	local floorHeight = self.height * self.tileH
	self.floorImage = love.graphics.newCanvas(floorWidth, floorHeight)

	-- Render floor to floor canvas
	self.floorImage:clear()
	self.floorImage:renderTo(function()	
		local defaultBlend = love.graphics.getBlendMode()
		love.graphics.setBlendMode("alpha")
		for i=0, self.width-1 do
			for j=0, self.height-1 do
				local tile = layer.tiles[i][j].char

				if(tile ~= " ") then
					local x, y = i*tileW, j*tileH
					love.graphics.draw(self.tileset, self.quads[tile], x, y)
				end
			end
		end

		-- Reset blend mode and canvas
		love.graphics.setBlendMode(defaultBlend)
	end)
end

function Map:generateSoftBricks(rate)
	local layer = self.layer[self.raycastLayer]

	for x=0, self.width-1 do
		for y=0, self.height-1 do
			local currentTile = layer.tiles[x][y]
			local rand = love.math.random()

			if(currentTile.char == " " and rand <= rate) then
				currentTile.char = "b"
				self:addAttributes(x, y, "collidable", "breakable")
			end
		end
	end

	-- Clear squares around spawn points
	for i=1,4 do
		local player = string.format("%d", i)
		local spawnTile = self.misc[player]
		local x, y = spawnTile.x, spawnTile.y
		local tiles = {
			{x = x, y = y},
			{x = x+1, y = y},
			{x = x-1, y = y},
			{x = x, y = y+1},
			{x = x, y = y-1}
		}

		for _, tile in ipairs(tiles) do
			if(not self:isOutOfBounds(tile.x, tile.y) and
			   layer.tiles[tile.x][tile.y].char == "b") then
				local currentTile = layer.tiles[tile.x][tile.y]

				currentTile.char = " "
				self:clearAttributes(tile.x, tile.y)
			end
		end
	end
end

function Map:drawMinimap(x, y, mode)
	local layer = self.layer[self.raycastLayer]
	local scale = self.minimapScale

	-- if mode is declared and is "center"
	if(mode == "center") then
		local mapW, mapH = self.width*scale, self.height*scale
		x, y = x-(mapW/2), y-(mapH/2)
	end

	-- save graphics settings just in case
	-- and then translate for easy drawing
	love.graphics.push()
	love.graphics.translate(x, y)
	-- save color for later
	local defaultColor = {love.graphics.getColor()}

	-- draw the raycast-able layer 
	for i=0, layer.width-1 do
		for j=0, layer.height-1 do
			local tile = layer.tiles[i][j].char
			local icon = self.minimapIcons[tile]

			if(tile ~= " ") then
				local m, n = i*scale, j*scale
				local color = {icon.r, icon.g, icon.b}

				love.graphics.setColor(color)
				love.graphics.rectangle("fill", m, n, scale, scale)
			end

			-- draw entities in tile
			local entities = self:getEntitiesAt(i, j)
			local entityCount = self:getEntityCount(i, j)
			if(entityCount > 0) then
				local d = findNextSquareBase(entityCount)
				local tileScale = scale / d
				local count = 0

				for _, entity in pairs(entities) do
					local m = (i*scale)+((count%d)*tileScale)+
							  (tileScale/2)
					local n = (j*scale)+(floor(count/d)*tileScale)+
							  (tileScale/2)

					love.graphics.setColor(entity.minimapIcon)
					love.graphics.circle("line", m, n, tileScale/2, 100)

					local tx = m+cos(entity.rotation)*(tileScale/2)
					local ty = n+sin(entity.rotation)*(tileScale/2)
					love.graphics.line(m, n, tx, ty)

					count = count + 1
				end
			end
		end
	end

	-- reset the color and reset graphics settings
	love.graphics.setColor(defaultColor)
	love.graphics.pop()
end

-- Entity functions
function Map:addEntityToTile(entity, x, y)
	local layer = self.layer[self.raycastLayer]
	local x, y = floor(x), floor(y)
	local n = self:getEntityCount(x, y)

	layer.tiles[x][y].entities[entity.id] = entity
	layer.tiles[x][y].entityCount = n + 1
end

function Map:removeEntityFromTile(entityID, x, y)
	local layer = self.layer[self.raycastLayer]
	local x, y = floor(x), floor(y)
	local n = self:getEntityCount(x, y)

	layer.tiles[x][y].entities[entityID] = nil
	layer.tiles[x][y].entityCount = n - 1
end

function Map:getEntitiesAt(x, y)
	local layer = self.layer[self.raycastLayer]
	local x, y = floor(x), floor(y)
	
	return layer.tiles[x][y].entities
end

function Map:getEntityCount(x, y)
	local layer = self.layer[self.raycastLayer]
	local x, y = floor(x), floor(y)
	
	return layer.tiles[x][y].entityCount
end

function Map:addEntity(entity)
	local x = floor(entity.position.x)
	local y = floor(entity.position.y)

	-- Add entity to table without breaking its order for ipairs to work
	table.insert(self.entities, entity)
	entity.map = self

	-- Set the entities id
	self.eid = self.eid + 1
	entity.id = self.eid

	-- Add to the tile
	self:addEntityToTile(entity, x, y)
end

function Map:removeEntity(entityID)
	local x, y

	local index = 1
	local size = #self.entities
	while index <= size do
		if(self.entities[index].id == entityID) then
			local entity = self.entities[index]
			x, y = entity.position:unpack()

			self.entities[index] = self.entities[size]
			self.entities[size] = nil
			size = size - 1
		else
			index = index + 1
		end
	end

	self:removeEntityFromTile(entityID, x, y)
end

function Map:updateEntities(dt)
	local layer = self.layer[self.raycastLayer]

	for _, entity in ipairs(self.entities) do
		local oldx = floor(entity.position.x)
		local oldy = floor(entity.position.y)
		entity:update(dt)
		local newx = floor(entity.position.x)
		local newy = floor(entity.position.y)

		if(oldx ~= newx or oldy ~= newy) then
			self:removeEntityFromTile(entity.id, oldx, oldy)
			self:addEntityToTile(entity, newx, newy)
		end
	end
end
-- End entity functions

-- Tile functions
function Map:isOutOfBounds(x, y)
	if(x < 0 or x >= self.width or
	   y < 0 or y >= self.height) then
		return true
	end

	return false
end

function Map:hasAttribute(x, y, attribute)
	if(self:isOutOfBounds(x, y)) then return false end
	x = floor(x)
	y = floor(y)

	local layer = self.layer[self.raycastLayer]
	local tile = layer.tiles[x][y]
	
	return tile.attributes[attribute] ~= nil
end

function Map:removeAttributes(x, y, ...)
	if(self:isOutOfBounds(x, y)) then return end
	x = floor(x)
	y = floor(y)

	local layer = self.layer[self.raycastLayer]
	local tile = layer.tiles[x][y]
	local args = { n = select("#", ...), ...}

	for i=1, args.n do
		if(args[i] ~= nil) then
			tile.attributes[args[i]] = nil
		end
	end
end

function Map:clearAttributes(x, y)
	if(self:isOutOfBounds(x, y)) then return end
	x = floor(x)
	y = floor(y)

	local layer = self.layer[self.raycastLayer]
	local tile = layer.tiles[x][y]

	for k in pairs(tile.attributes) do
		tile.attributes[k] = nil
	end
end

function Map:addAttributes(x, y, ...)
	if(self:isOutOfBounds(x, y)) then return end
	x = floor(x)
	y = floor(y)

	local layer = self.layer[self.raycastLayer]
	local tile = layer.tiles[x][y]
	local args = { n = select("#", ...), ...}

	for i=1, args.n do
		if(args[i] ~= nil) then
			tile.attributes[args[i]] = true
		end
	end
end

function Map:isBlocking(x, y)
	if(self:isOutOfBounds(x, y)) then return true end
	return self:hasAttribute(x, y, "collidable")
end
-- End tile functions

-- Delegated spawner functions
function Map:spawn(entity, time)
	self.spawner:spawn(entity, time)
end

function Map:spawnAll(...)
	self.spawner:spawnAll(...)
end

function Map:updateSpawner(dt)
	self.spawner:update(dt)
end
-- End delegated spawner functions

-- Delegated raycaster functions
function Map:castRays()
	self.raycaster:clearDrawables()
	self.raycaster:castRays()
	self.raycaster:spritecast()
end

function Map:drawRaycast()
	local sprites = 0
	self.raycaster:sortDrawables()
	for _, d in ipairs(self.raycaster.drawables) do
		if(d.type == "wall") then
			love.graphics.draw(self.tileset, d.quad,
							   d.x, d.y, 0, d.sx, d.sy)
		elseif(d.type == "sprite") then
			sprites = sprites + 1
			local entity = self.entities[d.id]
			if(entity.faces) then entity:face(d.face) end
			entity.spriteData = d
			entity:draw()
		end
	end
end

function Map:drawFloor()
	self.raycaster:floorcast()
end
-- End delegated raycaster functions

return Map
