local Vector = require("libs.hump.vector")
local Raycaster = Class("Raycaster")

local sqrt, ceil, floor = math.sqrt, math.ceil, math.floor
local asin, sin, cos = math.asin, math.sin, math.cos
local tan, atan2, acos = math.tan, math.atan2, math.acos
local pi, twoPi = math.pi, math.pi * 2

function Raycaster:init(map)
	-- map reference
	self.map = map
	-- canvas for floor-casting
	self.floor = nil
	-- use whose "eyes" for raycasting
	self.unit = nil

	-- raycast data
	self.fov = 0
	self.viewDist = 0
	self.rayCount = 0
	self.furthest = {
		wall = 0,
		strip = 0,
		top = 0,
		bottom = 0
	}
	self.slices = {
		wall = {},
		floor = {}
	}
	self.raycasts = {}
	self.drawables = {}
	self.sprites = {}
end

function Raycaster:initRaycaster()
	self.fov = math.rad(options.fov)
	self.viewDist = (options.width/2) / (tan(self.fov/2))
	self.rayCount = ceil(options.width/options.stripWidth)

	local map = self.map
	local wall = self.slices.wall
	local floor = self.slices.floor

	-- Pre-slice the tiles for raycasting
	for char, quad in pairs(map.quads) do
		local x, y, w, h = quad:getViewport()
		wall[char] = {}
		for i=0, map.tileW-1 do
			wall[char][i] = love.graphics.newQuad(x+i, y, 1, h,
												  map.refW,
												  map.refH)
		end
	end

	-- Pre-slice the floor for floor-casting
	local floorWidth, floorHeight = map.floorImage:getDimensions()
	for i=0, floorHeight-1 do
		floor[i] = love.graphics.newQuad(0, floorHeight-i,
										 floorWidth, 1,
										 floorWidth, floorHeight)
	end

	-- Create canvas for floor-casting
	self.floor = love.graphics.newCanvas(map.floorImage:getDimensions())
	self.floor:setFilter("nearest", "nearest")
end

function Raycaster:castRays()
	local unitPos = self.unit.position
	local unitRot = self.unit.rotation
	local map = self.map

	self.furthest.wall = 0
	for i=0, self.rayCount-1 do
		local rayScreenPos = (-self.rayCount/2 + i) * options.stripWidth
		local rayViewDist = sqrt(rayScreenPos*rayScreenPos +
								 self.viewDist*self.viewDist)
		local rayAngle = asin(rayScreenPos / rayViewDist)
		self:castSingleRay(i, unitRot + rayAngle)
	end
end

function Raycaster:castSingleRay(stripID, angle)
	local unitPos = self.unit.position
	local unitRot = self.unit.rotation
	local eyeHeight = self.unit.eyeHeight or 0.5
	local map = self.map
	local layer = map.layer[map.raycastLayer]

	angle = angle % twoPi
	if(angle < 0) then angle = angle + twoPi end

	local right = (angle > twoPi * 0.75) or (angle < twoPi * 0.25)
	local up = (angle < 0) or (angle > pi)

	local angleSin = sin(angle)
	local angleCos = cos(angle)
	if(angleSin == 0) then angleSin = 0.0001 end
	if(angleCos == 0) then angleCos = 0.0001 end

	local vHit = false
	local dist = 0
	local hit = Vector.new(0, 0)
	local wall = Vector.new(0, 0)
	local textureX

	-- Vertical check
	local slope = angleSin / angleCos
	local dx = right and 1 or -1
	local dy = dx * slope
	local x = right and ceil(unitPos.x) or floor(unitPos.x)
	local y = unitPos.y + (x - unitPos.x) * slope
	while(not map:isOutOfBounds(x, y)) do
		local wallX = right and floor(x) or floor(x - 1)
		local wallY = floor(y)
		if(map:isBlocking(wallX, wallY)) then
			local d = Vector.new(x, y) - unitPos
			dist = d:len()
			textureX = y % 1
			if(not right) then textureX = 1 - textureX end
			hit.x = x
			hit.y = y
			wall.x = wallX
			wall.y = wallY
			vHit = true
			break
		end
		x = x + dx
		y = y + dy
	end
	-- End vertical check
	-- Horizontal check
	local slope = angleCos / angleSin
	local dy = up and -1 or 1
	local dx = dy * slope
	local y = up and floor(unitPos.y) or ceil(unitPos.y)
	local x = unitPos.x + (y-unitPos.y) * slope
	while(not map:isOutOfBounds(x, y)) do
		local wallY = up and floor(y-1) or floor(y)
		local wallX = floor(x)
		if(map:isBlocking(wallX, wallY)) then
			local d = Vector.new(x, y) - unitPos
			local blockDist = d:len()
			if(dist == 0 or blockDist < dist) then
				dist = blockDist
				textureX = x % 1
				if(not up) then textureX = 1 - textureX end
				hit.x = x
				hit.y = y
				wall.x = wallX
				wall.y = wallY
				vHit = false
			end
			break
		end
		x = x + dx
		y = y + dy
	end
	-- End horizontal check

	if(dist > 0) then
		self.raycasts[stripID] = hit:clone()
		-- Correct fish-eye
		local origDist = dist
		dist = dist * cos(unitRot - angle)
		local height = self.viewDist / dist
		local texPos = floor(textureX * map.tileW)
		local tile = layer.tiles[wall.x][wall.y].char
		local sw, sh = options.width, options.height
		local stripWidth = options.stripWidth
		local scale = ((sh+height)/2-(sh-height)/2)/map.tileH
		-- Out of bounds for 90 FOV bug fix
		if(texPos >= map.tileW) then texPos = map.tileW-1 end
		self.drawables[#self.drawables + 1] = {
			type = "wall",
			x = stripID * stripWidth,
			y = (sh - height) * eyeHeight,
			z = floor(origDist * 100),
			sx = stripWidth,
			sy = scale,
			quad = self.slices.wall[tile][texPos]
		}
		-- stuff for floorcasting
		if(self.furthest.wall < dist) then
			self.furthest.wall = dist
			self.furthest.strip = stripID
			self.furthest.top = (sh-height)*eyeHeight
			self.furthest.bottom = self.furthest.top + scale * map.tileH
		end
	end
end

function Raycaster:spritecast()
	local unit = self.unit
	local map = self.map

	for i, entity in ipairs(map.entities) do
		-- Skip if entity is the raycaster unit or not visible to raycaster unit
		if(entity ~= unit and
		   (entity.visibleBy == nil or
		   	entity.visibleBy ~= nil and entity:visibleTo(unit.id))) then
			local delta = entity.position - unit.position
			local dist = delta:len()
			local angle = atan2(delta.y, delta.x) - unit.rotation
			local size = self.viewDist / (cos(angle) * dist)
			local x = (options.width / 2) + (tan(angle) * self.viewDist)
			local y = (options.height - size) * unit.eyeHeight
			local sx = size / map.spriteW
			local sy = size / map.spriteH
			local face = ""
			-- Angle between entity and unit
			local a = Vector.new(cos(angle + unit.rotation),
								 sin(angle + unit.rotation))
			-- Entity facing vector
			local b = Vector.new(cos(entity.rotation),
								 sin(entity.rotation))
			-- Unit facing vector
			local c = Vector.new(cos(unit.rotation),
								 sin(unit.rotation))
			-- Entity's left vector (since 270 is upward, not downward)
			local d = Vector.new(-sin(entity.rotation),
								  cos(entity.rotation))
			local theta = acos(a * b)
			local phi = acos(a * c)
			-- Skip if unit is not visible within FOV
			if((a * c) > 0 and phi <= self.fov*0.6) then
				-- Set entities to their proper facing
				if(entity.faces) then
					local isLeft = (d * delta < 0)
					if(theta > 0 and theta <= pi*0.25) then
						face = "back"
					elseif(theta > pi*0.25 and theta <= pi*0.75) then
						face = isLeft and "right" or "left"
					elseif(theta > pi*0.75 and theta <= pi) then
						face = "front"
					end
				end
				if(sx < 0) then sx = 0 end
				if(sy < 0) then sy = 0 end
				if(sx > 0 and sy > 0) then
					self.drawables[#self.drawables + 1] = {
						type = "sprite",
						id = i,
						x = x - size/2,
						y = y,
						z = floor(dist * 100),
						sx = sx,
						sy = sy,
						face = face
					}
				end
			end
		end
	end
end

function Raycaster:sortDrawables()
	table.sort(self.drawables, function(a, b)
		return a.z > b.z
	end)
end

function Raycaster:clearDrawables()
	for k in pairs(self.drawables) do
		self.drawables[k] = nil
	end
end

function Raycaster:floorcast()
	local unit = self.unit
	local map = self.map
	local unitPos = unit.position
	local unitRot = unit.rotation
	local fw, fh = self.floor:getDimensions()
	local sw, sh = options.width, options.height
	local furthest = self.furthest
	local rayViewDist = sqrt(furthest.wall*furthest.wall +
							 self.viewDist*self.viewDist)
	local rayAngle = asin(furthest.wall/rayViewDist)
	local cosAngle = cos(rayAngle)
	-- Prepare pre rotated image for floorcasting
	self.floor:clear()
	self.floor:renderTo(function()
		local width, height = self.floor:getDimensions()
		local defaultBlendMode = love.graphics.getBlendMode()
		love.graphics.setBlendMode("alpha")
		love.graphics.draw(map.floorImage, width/2, height,
						  -unitRot-pi/2, nil, nil,
						   unitPos.x * map.tileW,
						   unitPos.y * map.tileH)
		love.graphics.setBlendMode(defaultBlendMode)
	end)
	-- Draw floors
	for y=furthest.bottom, sh do
		if(cosAngle == 0) then cosAngle = 0.0001 end
		local dist = ((self.viewDist*((map.tileH*(1-unit.eyeHeight)))/
					 (y-sh*unit.eyeHeight))/cosAngle)/map.tileH
		local size = rayViewDist / dist / map.tileW
		local slice = floor(dist * map.tileH)
		if(slice >= fh) then slice = fh-1
		elseif(slice < 0) then slice = 0 end

		love.graphics.draw(self.floor, self.slices.floor[slice],
						  (sw/2), y, 0, size, 1, fw/2, 0)
	end
end

return Raycaster
