local ResourceHandler = Class("ResourceHandler")

local function buildPath(...)
	local args = {...}

	if(table.getn(args) <= 0) then return "" end

	if(table.getn(args) ~= 1) then
		local path = ""
		for _, v in ipairs(args) do
			if(v ~= "") then
				path = path .. "/" .. v
			end
		end
		
		return path:sub(2)
	else
		return args[1]
	end
end

function ResourceHandler:init(assetsFolder, otherFolders)
	self.root = assetsFolder or ""
	otherFolders = (type(otherFolders) == "table") and otherFolders or {}

	self.folders = {}
	self.folders.images = buildPath(self.root, otherFolders.images)
	self.folders.fonts = buildPath(self.root, otherFolders.fonts)

	self.loadedFonts = {}
	self:font(nil, "default")

	self.imageCache = {}
end

function ResourceHandler:path(type, ...)
	return buildPath(self.folders[type], ...)
end

function ResourceHandler:image(filename, colorKey, min, max)
	if(self.imageCache[filename]) then
		return self.imageCache[filename]
	else
		min = min or "nearest"
		max = max or min

		local image = love.graphics.newImage(buildPath(self.folders.images, filename))

		if(colorKey and type(colorKey) == "table") then
			local imageData = image:getData()
			imageData:mapPixel(function(x, y, r, g, b, a)
				if(colorKey[1] == r and colorKey[2] == g and colorKey[3] == b) then
					a = 0
				end
				return r, g, b, a
			end)
			image = love.graphics.newImage(imageData)
		end

		image:setFilter(min, max)

		self.imageCache[filename] = image
		return image
	end

end

function ResourceHandler:font(filename, fontName, fontSize)
	filename = filename and buildPath(self.folders.fonts, filename) or nil
	fontSize = fontSize or 12

	if(not self.loadedFonts[fontName]) then
		-- if(debug) then print(string.format("Fonts: Creating key for '%s'", fontName)) end
		self.loadedFonts[fontName] = {}
	else
		-- if(debug) then print(string.format("Fonts: Key '%s' already exists", fontName)) end
		if(self.loadedFonts[fontName][fontSize]) then
			-- if(debug) then print(string.format("Fonts: Size %d for '%s' is already loaded", fontSize, fontName)) end
			return
		end
	end

	self.loadedFonts[fontName][fontSize] = love.graphics.newFont(filename, fontSize)
end

function ResourceHandler:getFont(fontName, fontSize)
	fontName = fontName or "default"
	fontSize = fontSize or 12

	if(not self.loadedFonts[fontName] or not self.loadedFonts[fontName][fontSize]) then
		-- if(debug) then 
		-- 	print(string.format("Font Error: Key '%s' or Size %d for '%s' doesn't exist", fontName, fontSize, fontName))
		-- end
		return nil
	end

	return self.loadedFonts[fontName][fontSize]
end

function ResourceHandler:file(filename)
	return love.filesystem.read(buildPath(self.root, filename))
end

return ResourceHandler
